/************************************************************************
 *   IRC - Internet Relay Chat, dynconf.c
 *   (C) 1999 Carsten Munk (Techie/Stskeeps) <cmunk@toybox.flirt.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define DYNCONF_C
#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "userload.h"
#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef _WIN32
#include <utmp.h>
#else
#include <io.h>
#endif
#include <fcntl.h>
#include "h.h"

ID_CVS("$Id: dynconf.c,v 1.3 1999/11/07 13:12:16 cmunk Exp $");
ID_Copyright("(C) 1999 Carsten Munk");
#define DoDebug fprintf(stderr, "[%s] %s | %li\n", babuf, __FILE__, __LINE__);
#define AllocCpy(x,y) x = (char *) MyMalloc(strlen(y) + 1); strcpy(x,y)
#define XtndCpy(x,y) x = (char *) MyMalloc(strlen(y) + 2); *x = '\0'; strcat(x, "*"); strcpy(x,y)

aConfiguration iConf;
int	 icx = 0;
void iCstrip(char *line)
{
        char *c;

        if ((c = strchr(line, '\n')))   *c = '\0';
        if ((c = strchr(line, '\r')))   *c = '\0';
}

/*
        	for (q = p; *q; q++) {
                	printf("%x ", *q);
                }
*/

int		load_conf(char	*file)
{
	FILE	*zConf;
    char    *stat;
	char	*buf = MyMalloc(1024);
	char	*babuf = MyMalloc(1024);
	char	*p, *q;
	char	*var, *setto;          
    long    aint;

	zConf = fopen(file, "r");
	if (zConf == NULL) {
#ifdef _WIN32
        MessageBox(NULL, "UnrealIRCD/32 Init Error", "Unable to load dynamic config (or network) file !! ", MB_OK);
#else
		fprintf(stderr, "[error] Couldn't load %s !!!", file);
#endif        
		exit(-1);
	}
	*buf = '\1';
    stat = buf;
	while (stat != NULL)
	{
		stat = fgets(buf,1020, zConf);
		if (*buf == '#')       /* comment */
			continue;
		if (*buf == '/')       /* comment */
			continue;
		iCstrip(buf);    /* strip crlf .. */
		strcpy(babuf, buf);
		if (*buf == '\0')
			continue;
	      	p = strtok(buf, " ");
		if (strcmp(p, "Include")==0)
		{
	 			strtok(NULL, " ");
                setto = strtok(NULL, "");
#ifndef _WIN32
#endif
	 			load_conf(setto);
		}
		 else
		if (strcmp(p, "Set")==0)
		{
			var = strtok(NULL, " ");
			/* Yes it was a good idea when i moved it. */
			if (var == NULL)
				continue;
			if (*var == '\0')
				continue;

			strtok(NULL, " ");
			setto = strtok(NULL, "");
			if (setto == NULL)
				continue;
			if (*setto == '\0')
				continue;
			if (strlen(setto) == 1)
			{
				switch (*setto)
				{
					case '0' : aint = 0; break;
					case '1' : aint = 1; break;
#ifndef _WIN32
					default:   fprintf(stderr, "[error] Boolean field %s of %s is a non-boolean", var, file); 
#endif
							   aint = 0;
				}
	
				if (strcmp(var, "MODE_X")==0)
					MODE_X = aint;
				if (strcmp(var, "TRUEHUB")==0) {
					TRUEHUB = aint;
#ifndef HUB
					if (aint == 1) {
#ifdef _WIN32
				        MessageBox(NULL, "UnrealIRCD/32 Init Error", "I'm a leaf NOT an hub! ", MB_OK);
												
#else
						fprintf(stderr, "[error] I'm a leaf NOT a hub!\n");
#endif
						return 0;
					}
#endif
				}
				if (strcmp(var, "CONFIG_FILE_STOP")==0) {
					if (aint == 1) {
#ifdef _WIN32
				        MessageBox(NULL, "UnrealIRCD/32 Init Error", "Read Config stop code in file", MB_OK);											
#else
						fprintf(stderr, "[fatal error] File stop code recieved in %s - RTFM\n", file);
#endif
						exit(-1);
					}
				}
#ifndef MegaIRC
				if (strcmp(var, "STOPSE")==0) {
					if (aint == 1) {
						fprintf(stderr, "[FATAL ERROR] This network file (%s) requires Unreal-MegaIRC[SE]\n", file);
						fprintf(stderr, "[-----------] Please use the special edition for this network\n");
						exit(-1);
					}
				}
#endif
				if (strcmp(var, "SHOWOPERS")==0) {
					SHOWOPERS = aint;
				}
				if (strcmp(var, "SHOWOPERMOTD")==0) {
					SHOWOPERMOTD = aint;
				}
       	     	if (strcmp(var, "iNAH")==0) {
       	     			iNAH = aint;
       			}
             	if (strcmp(var, "KILLDIFF")==0) {
             			KILLDIFF = aint;
             	}
                		continue;
        	 }
/*			 if (strcmp(var, "DOMAINNAME")==0) {
			 	AllocCpy(DOMAINNAME, setto);
			 	XtndCpy(DOMAINNAMEMASK, setto); 
			 }
*/
			 /* uhm networks */
			 if (strcmp(var, "ircnetwork")==0) {
			 	AllocCpy(ircnetwork, setto);
			 } 
			 	else
			 if (strcmp(var, "defserv")==0) {
			 	AllocCpy(defserv, setto);
			 }
			 	else
			 if (strcmp(var, "SERVICES_NAME")==0) {
			 	AllocCpy(SERVICES_NAME, setto);
			 }
			 	else
			 if (strcmp(var, "oper_host")==0) {
			 	AllocCpy(oper_host, setto);
			 }
				else
			 if (strcmp(var, "admin_host")==0) {
			 	AllocCpy(admin_host, setto);
			 }
				else
			 if (strcmp(var, "locop_host")==0) {
			 	AllocCpy(locop_host, setto);
			 }
			 	else
			 if (strcmp(var, "sadmin_host")==0) {
			 	AllocCpy(sadmin_host, setto);
			 }
				else
			 if (strcmp(var, "netadmin_host")==0) {
			 	AllocCpy(netadmin_host, setto);
			 }			 
             	else
			 if (strcmp(var, "techadmin_host")==0) {
			 	AllocCpy(techadmin_host, setto);
			 }			 
             	else
             if (strcmp(var, "coadmin_host")==0) {
             	AllocCpy(coadmin_host, setto);
             }
             	else
             if (strcmp(var, "hidden_host")==0) {
             	AllocCpy(hidden_host, setto);
             }
             	else
             if (strcmp(var, "netdomain")==0) {
             	AllocCpy(netdomain, setto);
             }
				else
			 if (strcmp(var, "helpchan")==0) {
			 	AllocCpy(helpchan, setto);
			 }
			 	else
			 if (strcmp(var, "STATS_SERVER")==0) {
			 	AllocCpy(STATS_SERVER, setto);
			 }
			 	else
			 if (strcmp(var, "net_quit")==0) {
			 	AllocCpy(net_quit, setto);
			 }
				else
			if (strcmp(var, "KLINE_ADDRESS")==0) {
				AllocCpy(KLINE_ADDRESS, setto);
			}
         }
	}
#ifndef _WIN32
	fprintf(stderr, "* Loaded %s ..\n", file);
#endif	
}

void	doneconf(void) {
	
}